#if 0
C -o charset main.o charsets.o
Exit 0
#endif
/*
 charsets.c (was objects.c at one point)

  Created: 2010.10.25
  Revised: 2010.12.23

2010.12.17 - after implementing map2list, added MacRoman reader, then CP1250.
2010.12.18 - fully implemented and tested UTF-8 reader
2010.12.20 - moved transcoder here, rewriting it so it's not one char at a time.
	expanded it to do input and output routines instead of just UTF-8.
	Fixed writeCharUTF_8's output mask.
2010.12.21 - other bug fixes.
2010.12.22 - added nb (number of bytes) to ftranscode.
	ASCII now returns low 7 bits instead of UNKNOWN when bit #7 is one.
	__UTF32_user8bit is now UTF16 (8 bytes, non-BMP surrogates allowed)
	since languages like Klingon are in no way standard.
2010.12.23 - started writers as callback to readers.
	Removed redundant makeCharUTF16.
	UTF16 surrogates in UTF32 code go to UNKNOWN.
	Wrote BMP scanner.
*/
/*
Encoding names/types source: character-sets.txt (2010.09.15)

	Assigned MIB enum Numbers
	-------------------------
	0-2		Reserved
	3-999		Set By Standards Organizations
	1000-1999	Unicode / 10646
	2000-2999	Vendor

  Aliases that start with "cs" have been added for use with the
IANA-CHARSET-MIB as originally defined in RFC3808, and as currently
maintained by IANA at iana.org/assignments/ianacharset-mib.

  For ISO-10646:
   The only difference between UTF and UCS-4/UCS-2 is the upper range (non-BMP).
   UTF-16 carries two surrogate ranges not handled by UCS-2: [D800,DC00,DFFF].

  5 and 6 byte UTF-8 sequences are invalid in the Unicode standard because the
smallest length for the highest value (0x10FFFF) is 4 bytes.
*/

/*
  The full 12 bit index range was reduced to 8 bits by eliminating undefined slots.
   Conversion C code:
	if (x!=0) if (x>=137) if (x>=246) x+= 2004; else x+= 1863;
	else if (x>=116) x+= 884; else if (x>=105) x+= 4; else x+= 2;
*/

#include <stdio.h>

/* optimizations (automatic zero extension is native for PowerPC) */
#if powerc
 #define uchar			UTF32
 #define uword			UTF32
#else
 #define uchar			UTF8
 #define uword			UTF16
#endif

/* ---- basic info functions */

#include "charsets.h"

#define IRNo \
	"\0\0\0\6\144\145\155\156\220\177\176\212\224\235\216\0\0\0\0\0" \
	"\4\13\17\21\25\74\105\0\0\0\2\0\0\0\0\12" \
	"\225\0\0\0\0\15\16\20\22\23\31\33\45\52\57\61" \
	"\62\63\66\67\71\72\75\106\124\125\126\127\130\131\0\133" \
	"\134\135\136\137\140\142\143\146\147\157\171\172\173\0\0\200" \
	"\0\0\213\215\217\222\223\226\227\230\231\232\233\236\237\0" \
	"\0\0\0\0\0\0\0\0\0\0\307\0\342"

int getMIB_IR_number(unsigned x)
{
	const uint8	*s = (uint8*)IRNo;

	if (x>=31)
		if (x<=34)
			return ((x-31)&-2)*500+(8002-(x&1));
		else if (x>=sizeof(IRNo))
			return 0;

	return s[x];
}

const char *getMIB_kind(unsigned x)
{
	if (x<=999)
		switch(x)
		{ /* unsure about 72, 81/82 */
		case 3: goto unitedstates;
		case 4: goto latin1;
		case 5: goto latin2;
		case 6: return "Latin-3";
		case 7: return "Latin-4";
		case 8: case 48: case 53: case 77: goto cyrillic;
		case 9: return "Latin-Arabic";
		case 10: case 55: case 45: case 47: return "Latin-Greek";
		case 11: goto latinhebrew;
		case 12: goto latin5;
		case 13: return "Latin-6";
		case 14: return "TextComm";
		case 15: return "JIS/Roman";
		case 16: case 17: case 63: case 98: return "JIS X 02xx";
		case 18: case 19: return "EUC-JIS"; /* Extended UNIX Code */
		case 20: goto unitedkingdom;
		case 21: return "Swedish for names";
		case 22: case 82: goto italian;
		case 23: case 81: goto spanish;
		case 24: goto german;
		case 25: goto danishnorwegian;
		case 26: case 46: goto french;
		case 27: case 103: case 106: goto unicode;
		case 28: return "ISO-646 basic";
		case 29: return "none";
		case 30: return "ISO-646 irv";
	/*	case 31: case 32: return "SEFI"; ??? */
		case 33: case 34: goto danonorwegian;
		case 35: goto swedish;
		case 36: case 37: case 38: case 102: return "Korean";
		case 39: case 40: case 42: goto japanese;
		case 41: case 72: goto katakana;
		case 43: goto portuguese;
		case 44: return "Greek-7-Old";
		case 49: return "JIS";
		case 50: return "View-Data";
	/*	case 51: case 52: case 66: return "unspecified";*/
		case 54: return "Cyrillic-1981";
		case 56: case 57: case 104: case 105: case 114: return "Chinese";
		case 58: return "Norwegian-2";
		case 59: return "Videotex-Supp1";
		case 60: return "Portuguese-2";
		case 61: return "Spanish-2";
		case 62: return "Hungarian";
		case 64: return "Greek-7";
		case 65: return "Arabic-7";
		case 67: case 68: case 69: return "Japanese OCR";
		case 70: case 71: return "Japanese HAND";
		case 73: return "E-13-B";
		case 74: return "ANSI X3.110-1983";
		case 75: case 76: return "T.61";
		case 78: return "Canadian-1";
		case 79: return "Canadian-2";
		case 80: return "Greek-1985";
	//	case 83: return "Greek-2";
		case 87: return "JS/YU";
		case 89: return "Serbian";
		case 90: return "Macedonian";
		case 91: return "Greek-CCITT";
		case 92: return "Cuba";
		case 94: return "ST SEV 358-88";
		case 95: return "Latin-1-2-5 Supp";
		case 96: return "Box";
		case 97: return "Lap";
		case 99: return "Danish";
		case 100: case 101: return "US/DK";
		case 110: return "Celtic"; /* Latin-8 */
		case 111: return "Latin-9";
		case 112: return "Latin-10";
		case 113: return "Chinese Windows";
		case 115: case 116: case 117: goto ebcdic;
	//	case 119: return "Kazak";
		}
	else if (x<=1999)
	{
		switch(x)
		{
		case 1000: return "Basic Multilingual 16-bit";
		case 1001: return "Multilingual 31-bit";
		case 1002: return "ASCII subset";
		case 1003: return "Latin Basic/1";
	/*	case 1004: */ japanese: return "Japanese";
		case 1005: return "presentation Latin-1-2-5"; /* GCSGID 1261 */
		case 1006: return "presentation Latin-4"; /* GCSGID 1268 */
		case 1007: return "presentation Cyrillic-Greek";
		case 1008: return "presentation Arabic";
		case 1009: return "presentation Hebrew";
		}
	unicode:
		return "Unicode";
	} else if (x<=2199)
		switch(x)
		{
		case 2000: case 2001: latin1: return "Latin-1";
		case 2002: latin2: return "Latin-2";
		case 2003: latin5: return "Latin-5";
		case 2004: return "LaserJet IIP-1989";
		case 2005: return "PostScript";
		case 2006: case 2007: return "ASCII+"; /* PCL 14J,13J */
		case 2008: return "VAX/VMS-1986";
		case 2009: return "Multilingual";
		case 2012: case 2067: case 2068:
		danishnorwegian: return "Danish-Norwegian"; /* PCL 11U */
		case 2013: latinhebrew: return "Latin-Hebrew";
		case 2014: case 2061: turkish: return "Turkish";
		case 2015: return "Symbols";
		case 2016: thai: return "Thai";
		case 2017: case 2018: case 2019: /* Oct 1992; 1U,15U,8M */
		case 2021: case 2022: case 2023: /* Oct 1992; 7J,6M,6J */
			return "Hewlett-Packard PCL 5";
		case 2024: return "Windows Japanese";
		case 2025: return "ASCII/PRC Kanji";
		case 2026: return "Taiwan Multi-byte";
		case 2027: return "MacRoman";
		case 2028: case 2029: case 2031:
		case 2032: case 2036: case 2042:
		ebcdic: return "EBCDIC"; /* Mar 1990 */
		case 2033: danonorwegian: return "Dano-Norwegian";
		case 2034: case 2069: case 2070:
		swedish: return "Swedish";
		case 2035: case 2072: italian: return "Italian";
		case 2037: case 2074: case 2075:
		case 2076: spanish: return "Spanish";
		case 2038: case 2077: unitedkingdom: return "United Kingdom";
		case 2039: katakana: return "Katakana";
		case 2040: case 2071: french: return "French";
		case 2041: case 2053: return "Arabic-1";
		case 2043: return "Hebrew";
		case 2044: case 2046: case 2047:
		case 2048: case 2050: case 2051:
		case 2058: case 2059: case 2060:
		case 2063:
			return "IBM Code";
		case 2045: return "BE/CH";
		case 2049: case 2056: return "Israel"; /* unsure */
		case 2052: return "DOS 3.3";
		case 2054: return "Greek";
		case 2055: return "ROECE/YU";
		case 2057: case 2084: cyrillic: return "Cyrillic";
		case 2062: return "Arabic-2";
		case 2064: case 2065: german: return "German";
		case 2066: return "Canadian French";
		case 2073: portuguese: return "Portuguese";
		case 2078: unitedstates: return "United States";
		case 2080: case 2081: return "Mnemonic+ASCII";
		case 2082: case 2083: return "Vietnamese";
		case 2085: return "Chinese+ASCII";
		case 2086: return "NLDG";
		case 2087: return "Baltic";
		case 2089: return "Multilingual+euro";
		case 2090: return "Latin-9+euro";
		case 2091: return "US+euro";
		case 2092: return "German+euro";
		case 2093: return "Dano-Norwegian+euro";
		case 2094: return "Swedish+euro";
		case 2095: return "Italian+euro";
		case 2096: return "Spanish+euro";
		case 2097: return "British+euro";
		case 2098: return "French+euro";
		case 2099: return "Intl.+euro";
		case 2100: return "Israel+euro";
		case 2101: return "Latin-1/Open Systems";
		case 2102: return "Taiwanese";
		case 2103: return "Cyrillic-Asian";
		case 2104: return "Amiga";
		}
	else if (x>=2250 && x<=2258)
		return "Windows";
	else if (x==2259)
		goto thai;

	return "?";
}

/* returns MIB enum no. */
const int getEncodingByIndex(uint8 x,
	const char **n,
	utf_32_reader *r)
{
	utf_32_reader	p = 0;
	const char	*s = 0;
	int		i = -1; /* blank */

	switch(x)
	{
	case 0: i = 0; break; /* default */
	case 1: s = "ASCII", i = 3, p = readUTF32_ASCII; break; /*1968; cp367*/
	case 2: s = "ISO-8859-1", i = 4, p = readUTF32_ISO8859_1; break; /*1987; cp819*/
	case 3: s = "ISO-8859-2", i = 5; break; /* 1987 */
	case 4: s = "ISO-8859-3", i = 6; break; /* 1988 */
	case 5: s = "ISO-8859-4", i = 7; break; /* 1988 */
	case 6: s = "ISO-8859-5", i = 8; break; /* 1988 */
	case 7: s = "ISO-8859-6", i = 9; break; /* 1987; ASMO-708 */
	case 8: s = "ISO-8859-7", i = 10; break; /* 1987; ELOT_928 */
	case 9: s = "ISO-8859-8", i = 11; break; /* 1988 */
	case 10: s = "ISO-8859-9", i = 12; break; /* 1989 */
	case 11: s = "ISO-8859-10", i = 13; break; /* 1992 */
	case 12: s = "ISO_6937-2-add", i = 14; break; /* 1983 */
	case 13: s = "JIS_X0201", i = 15; break; /* 1-byte only half-width Katakana */
	case 14: s = "JIS_Encoding", i = 16; break; /* 1991; ISO 2022 and escape sequences */
	case 15: s = "Shift_JIS", i = 17; break; /* 1997 JIS X0201,X0208 */
	case 16: s = "EUC-JP", i = 18; break; /* 4 UNIX sets 1990; ASCII,X0208,X0201,X0212 */
	case 17: s = "csEUCFixWidJapanese", i = 19; break; /* unpacked EUC-JP, 2 octets */
	case 18: s = "BS_4730", i = 20; break;
	case 19: s = "SEN_850200_C", i = 21; break;
	case 20: s = "IT", i = 22; break;
	case 21: s = "ES", i = 23; break;
	case 22: s = "DIN_66003", i = 24; break;
	case 23: s = "NS_4551-1", i = 25; break; /* ir60 */
	case 24: s = "NF_Z_62-010", i = 26; break;
	case 25: s = "ISO-10646-UTF-1", i = 27; break;
	case 26: s = "ref", i = 28; break; /* 1983 */
	case 27: s = "INVARIANT", i = 29; break;
	case 28: s = "iso-ir-2", i = 30; break; /* 1983 */
	case 29: s = "NATS-SEFI", i = 31; break;
	case 30: s = "NATS-SEFI-ADD", i = 32; break;
	case 31: s = "NATS-DANO", i = 33; break;
	case 32: s = "NATS-DANO-ADD", i = 34; break;
	case 33: s = "SEN_850200_B", i = 35; break;
	case 34: s = "KS_C_5601-1987", i = 36; break;
	case 35: s = "ISO-2022-KR", i = 37; break;
	case 36: s = "EUC-KR", i = 38; break;
	case 37: s = "ISO-2022-JP", i = 39; break;
	case 38: s = "ISO-2022-JP-2", i = 40; break;
	case 39: s = "JIS_C6220-1969-jp", i = 41; break;
	case 40: s = "JIS_C6220-1969-ro", i = 42; break;
	case 41: s = "PT", i = 43; break;
	case 42: s = "greek7-old", i = 44; break;
	case 43: s = "latin-greek", i = 45; break;
	case 44: s = "NF_Z_62-010_(1973)", i = 46; break;
	case 45: s = "Latin-greek-1", i = 47; break;
	case 46: s = "ISO_5427", i = 48; break;
	case 47: s = "JIS_C6226-1978", i = 49; break;
	case 48: s = "BS_viewdata", i = 50; break;
	case 49: s = "INIS", i = 51; break;
	case 50: s = "INIS-8", i = 52; break;
	case 51: s = "INIS-cyrillic", i = 53; break;
	case 52: s = "iso-ir-54", i = 54; break; /* ISO_5427:1981 */
	case 53: s = "iso-ir-55", i = 55; break; /* ISO_5428:1980 */
	case 54: s = "GB_1988-80", i = 56; break;
	case 55: s = "GB_2312-80", i = 57; break;
	case 56: s = "NS_4551-2", i = 58; break;
	case 57: s = "videotex-suppl", i = 59; break;
	case 58: s = "PT2", i = 60; break;
	case 59: s = "ES2", i = 61; break;
	case 60: s = "MSZ_7795.3", i = 62; break;
	case 61: s = "JIS_C6226-1983", i = 63; break;
	case 62: s = "greek7", i = 64; break;
	case 63: s = "ASMO_449", i = 65; break;
	case 64: s = "iso-ir-90", i = 66; break;
	case 65: s = "JIS_C6229-1984-a", i = 67; break;
	case 66: s = "JIS_C6229-1984-b", i = 68; break;
	case 67: s = "JIS_C6229-1984-b-add", i = 69; break;
	case 68: s = "JIS_C6229-1984-hand", i = 70; break;
	case 69: s = "JIS_C6229-1984-hand-add", i = 71; break;
	case 70: s = "JIS_C6229-1984-kana", i = 72; break;
	case 71: s = "ISO_2033-1983", i = 73; break;
	case 72: s = "NAPLPS", i = 74; break;
	case 73: s = "T.61-7bit", i = 75; break;
	case 74: s = "T.61-8bit", i = 76; break;
	case 75: s = "ECMA-cyrillic", i = 77; break;
	case 76: s = "CSA_Z243.4-1985-1", i = 78; break;
	case 77: s = "CSA_Z243.4-1985-2", i = 79; break;
	case 78: s = "CSA_Z243.4-1985-gr", i = 80; break;
	case 79: s = "ISO-8859-6-E", i = 81; break;
	case 80: s = "ISO-8859-6-I", i = 82; break;
	case 81: s = "T.101-G2", i = 83; break;
	case 82: s = "ISO-8859-8-E", i = 84; break;
	case 83: s = "ISO-8859-8-I", i = 85; break;
	case 84: s = "CSN_369103", i = 86; break;
	case 85: s = "JUS_I.B1.002", i = 87; break;
	case 86: s = "IEC_P27-1", i = 88; break;
	case 87: s = "JUS_I.B1.003-serb", i = 89; break;
	case 88: s = "JUS_I.B1.003-mac", i = 90; break;
	case 89: s = "greek-ccitt", i = 91; break;
	case 90: s = "NC_NC00-10:81", i = 92; break;
	case 91: s = "ISO_6937-2-25", i = 93; break;
	case 92: s = "GOST_19768-74", i = 94; break;
	case 93: s = "ISO_8859-supp", i = 95; break;
	case 94: s = "ISO_10367-box", i = 96; break;
	case 95: s = "latin-lap", i = 97; break;
	case 96: s = "JIS_X0212-1990", i = 98; break;
	case 97: s = "DS_2089", i = 99; break;
	case 98: s = "us-dk", i = 100; break;
	case 99: s = "dk-us", i = 101; break;
	case 100: s = "KSC5636", i = 102; break;
	case 101: s = "UNICODE-1-1-UTF-7", i = 103; break;
	case 102: s = "ISO-2022-CN", i = 104; break;
	case 103: s = "ISO-2022-CN-EXT", i = 105; break;
	case 104: s = "UTF-8", i = 106; break;
	case 105: s = "ISO-8859-13", i = 109; break; /* no alias */
	case 106: s = "ISO-8859-14", i = 110; break;
	case 107: s = "ISO-8859-15", i = 111; break;
	case 108: s = "ISO-8859-16", i = 112; break;
	case 109: s = "GBK", i = 113; break;
	case 110: s = "GB18030", i = 114; break;
	case 111: s = "OSD_EBCDIC_DF04_15", i = 115; break;
	case 112: s = "OSD_EBCDIC_DF03_IRV", i = 116; break;
	case 113: s = "OSD_EBCDIC_DF04_1", i = 117; break;
	case 114: s = "ISO-11548-1", i = 118; break;
	case 115: s = "KZ-1048", i = 119; break;
	case 116: s = "ISO-10646-UCS-2", i = 1000; break;
	case 117: s = "ISO-10646-UCS-4", i = 1001; break;
	case 118: s = "ISO-10646-UCS-Basic", i = 1002; break;
	case 119: s = "ISO-10646-Unicode-Latin1", i = 1003; break;
//	case 120: s = "ISO-10646-J-1", i = 1004; break;
	case 121: s = "ISO-Unicode-IBM-1261", i = 1005; break;
	case 122: s = "ISO-Unicode-IBM-1268", i = 1006; break;
	case 123: s = "ISO-Unicode-IBM-1276", i = 1007; break;
	case 124: s = "ISO-Unicode-IBM-1264", i = 1008; break;
	case 125: s = "ISO-Unicode-IBM-1265", i = 1009; break;
	case 126: s = "UNICODE-1-1", i = 1010; break;
	case 127: s = "SCSU", i = 1011; break;
	case 128: s = "UTF-7", i = 1012; break;
	case 129: s = "UTF-16BE", i = 1013, p = readUTF32_UTF16BE; break;
	case 130: s = "UTF-16LE", i = 1014, p = readUTF32_UTF16LE; break;
	case 131: s = "UTF-16", i = 1015; break;
	case 132: s = "CESU-8", i = 1016; break;
	case 133: s = "UTF-32", i = 1017; break;
	case 134: s = "UTF-32BE", i = 1018, p = readUTF32_UTF32BE; break;
	case 135: s = "UTF-32LE", i = 1019, p = readUTF32_UTF32LE; break;
	case 136: s = "BOCU-1", i = 1020; break;
	case 137: s = "ISO-8859-1-Windows-3.0-Latin-1", i = 2000; break; /* 1996 */
	case 138: s = "ISO-8859-1-Windows-3.1-Latin-1", i = 2001; break;
	case 139: s = "ISO-8859-2-Windows-Latin-2", i = 2002; break;
	case 140: s = "ISO-8859-9-Windows-Latin-5", i = 2003; break;
	case 141: s = "hp-roman8", i = 2004; break;
	case 142: s = "Adobe-Standard-Encoding", i = 2005; break;
	case 143: s = "Ventura-US", i = 2006; break;
	case 144: s = "Ventura-International", i = 2007; break;
	case 145: s = "DEC-MCS", i = 2008; break;
	case 146: s = "IBM850", i = 2009; break;
	case 147: s = "IBM852", i = 2010; break; /*these two:NLS RM Vol2 SE09-8002-01, Mar1990*/
	case 148: s = "IBM437", i = 2011; break;
	case 149: s = "PC8-Danish-Norwegian", i = 2012; break;
	case 150: s = "IBM862", i = 2013; break;
	case 151: s = "PC8-Turkish", i = 2014; break;
	case 152: s = "IBM-Symbols", i = 2015; break;
	case 153: s = "IBM-Thai", i = 2016; break;
	case 154: s = "HP-Legal", i = 2017; break;
	case 155: s = "HP-Pi-font", i = 2018; break;
	case 156: s = "HP-Math8", i = 2019; break;
	case 157: s = "Adobe-Symbol-Encoding", i = 2020; break;
	case 158: s = "HP-DeskTop", i = 2021; break;
	case 159: s = "Ventura-Math", i = 2022; break;
	case 160: s = "Microsoft-Publishing", i = 2023; break;
	case 161: s = "Windows-31J", i = 2024; break; /* further ext. of Shift_JIS */
	case 162: s = "GB2312", i = 2025; break; /* People's Republic of China */
	case 163: s = "Big5", i = 2026; break;
	case 164: s = "macintosh", i = 2027, p = readUTF32_MacRoman; break;
	case 165: s = "IBM037", i = 2028; break;
	case 166: s = "IBM038", i = 2029; break;
	case 167: s = "IBM273", i = 2030; break;
	case 168: s = "IBM274", i = 2031; break;
	case 169: s = "IBM275", i = 2032; break;
	case 170: s = "IBM277", i = 2033; break;
	case 171: s = "IBM278", i = 2034; break;
	case 172: s = "IBM280", i = 2035; break;
	case 173: s = "IBM281", i = 2036; break;
	case 174: s = "IBM284", i = 2037; break;
	case 175: s = "IBM285", i = 2038; break;
	case 176: s = "IBM290", i = 2039; break;
	case 177: s = "IBM297", i = 2040; break;
	case 178: s = "IBM420", i = 2041; break;
	case 179: s = "IBM423", i = 2042; break;
	case 180: s = "IBM424", i = 2043; break;
	case 181: s = "IBM500", i = 2044; break; /* PC8 */
	case 182: s = "IBM851", i = 2045; break;
	case 183: s = "IBM855", i = 2046; break;
	case 184: s = "IBM857", i = 2047; break;
	case 185: s = "IBM860", i = 2048; break;
	case 186: s = "IBM861", i = 2049; break;
	case 187: s = "IBM863", i = 2050; break;
	case 188: s = "IBM864", i = 2051; break;
	case 189: s = "IBM865", i = 2052; break;
	case 190: s = "IBM868", i = 2053; break;
	case 191: s = "IBM869", i = 2054; break;
	case 192: s = "IBM870", i = 2055; break;
	case 193: s = "IBM871", i = 2056; break;
	case 194: s = "IBM880", i = 2057; break;
	case 195: s = "IBM891", i = 2058; break;
	case 196: s = "IBM903", i = 2059; break;
	case 197: s = "IBM904", i = 2060; break;
	case 198: s = "IBM905", i = 2061; break;
	case 199: s = "IBM918", i = 2062; break;
	case 200: s = "IBM1026", i = 2063; break;
	case 201: s = "EBCDIC-AT-DE", i = 2064; break;
	case 202: s = "EBCDIC-AT-DE-A", i = 2065; break;
	case 203: s = "EBCDIC-CA-FR", i = 2066; break;
	case 204: s = "EBCDIC-DK-NO", i = 2067; break;
	case 205: s = "EBCDIC-DK-NO-A", i = 2068; break;
	case 206: s = "EBCDIC-FI-SE", i = 2069; break;
	case 207: s = "EBCDIC-FI-SE-A", i = 2070; break;
	case 208: s = "EBCDIC-FR", i = 2071; break;
	case 209: s = "EBCDIC-IT", i = 2072; break;
	case 210: s = "EBCDIC-PT", i = 2073; break;
	case 211: s = "EBCDIC-ES", i = 2074; break;
	case 212: s = "EBCDIC-ES-A", i = 2075; break;
	case 213: s = "EBCDIC-ES-S", i = 2076; break;
	case 214: s = "EBCDIC-UK", i = 2077; break;
	case 215: s = "EBCDIC-US", i = 2078; break;
	case 216: s = "UNKNOWN-8BIT", i = 2079; break;
	case 217: s = "MNEMONIC", i = 2080; break;
	case 218: s = "MNEM", i = 2081; break;
	case 219: s = "VISCII", i = 2082; break;
	case 220: s = "VIQR", i = 2083; break;
	case 221: s = "KOI8-R", i = 2084; break;
	case 222: s = "HZ-GB-2312", i = 2085; break; /* 1995 */
	case 223: s = "IBM866", i = 2086; break; /* 1994 */
	case 224: s = "IBM775", i = 2087; break;
	case 225: s = "KOI8-U", i = 2088; break;
	case 226: s = "IBM00858", i = 2089; break;
	case 227: s = "IBM00924", i = 2090; break;
	case 228: s = "IBM01140", i = 2091; break;
	case 229: s = "IBM01141", i = 2092; break;
	case 230: s = "IBM01142", i = 2093; break;
	case 231: s = "IBM01143", i = 2094; break;
	case 232: s = "IBM01144", i = 2095; break;
	case 233: s = "IBM01145", i = 2096; break;
	case 234: s = "IBM01146", i = 2097; break;
	case 235: s = "IBM01147", i = 2098; break;
	case 236: s = "IBM01148", i = 2099; break;
	case 237: s = "IBM01149", i = 2100; break;
	case 238: s = "Big5-HKSCS", i = 2101; break;
	case 239: s = "IBM1047", i = 2102; break;
	case 240: s = "PTCP154", i = 2103; break;
	case 241: s = "Amiga-1251", i = 2104; break; /* 2004 */
	case 242: s = "KOI7-switched", i = 2105; break;
	case 243: s = "BRF", i = 2106; break;
	case 244: s = "TSCII", i = 2107; break;
	case 245: s = "CP51932", i = 2108; break;
	case 246: s = "windows-1250", i = 2250, p = readUTF32_WCP1250; break;
	case 247: s = "windows-1251", i = 2251; break;
	case 248: s = "windows-1252", i = 2252, p = readUTF32_WCP1252; break;
	case 249: s = "windows-1253", i = 2253; break;
	case 250: s = "windows-1254", i = 2254; break;
	case 251: s = "windows-1255", i = 2255; break;
	case 252: s = "windows-1256", i = 2256; break;
	case 253: s = "windows-1257", i = 2257; break;
	case 254: s = "windows-1258", i = 2258; break;
	case 255: s = "TIS-620", i = 2259;
	}
end:
	if (n)
		*n = s;
	if (r)
		*r = p;

	return i;
}

/* ---- BMP scanner */

static unsigned BMPwchr(UTF32 i,
	const UTF16 *t,
	UTF32 s,
	int c)
{
	if (i<s)
		return i;
	if (c>0 && i>0xFFFF)
	{
		const UTF16	*e = t+c;

		do
			if ((uword)i==*t++)
				return e-t+s+1;
		while(t<e);
	}

	return 0x3F; /* ? */
}

/* ---- input-to-UTF-32 routines */

__cb_utf32reader(readUTF32_ASCII)
{
	return *(*s)++&0x7F;
}

__cb_utf32reader(readUTF32_ISO8859_1)
{
	return *(*s)++;
}

__cb_utf32reader(readUTF32_GSM0338) /* ETSI */
{
	const uint8	*p = *s;
	UTF32		c;

	if ((c=*p++)==0)
		c = (*p>12)<<6;
	else if ((c-=0x1B)==0)
	{
		const uint8	*esclow = (const uint8*)
			"{}\0\0\0\0\0\\\0\0\0\0\0\0\0\0\0\0\0\0[~\0|";
		if ((c=*p++-0x65)==0)
			c = 0x20AC;
		else if ((c-=10-'e')==0)
			c = 12;
		else if ((c-=20-10)==0)
			c = 94;
		else if ((c-=40-20)>sizeof(esclow) || (c=esclow[c])==0)
			c = 0xA0, --p;
	} else
	{
		const uint8	*GSMtab = (const uint8*)
			"\xA3$\xA5\xE8\xE9\xF9\xEC\xF2\xE7\12\xD8\xF8\15\xC5\xE5"
			"\4_\26\3\13\31\20\30\36\xC6\xE6\xDF\xC9"
			" !\"#\xA4%&'()*+,-./"
			"0123456789:;<=>?"
			"\xA1ABCDEFGHIJKLMNO"
			"PQRSTUVWXYZ\xC4\xD6\xD1\xDC\xA7"
			"\xBFabcdefghijklmno"
			"pqrstuvwxyz\xE4\xF6\xF1\xFC\xE0";
		if ((c+=0x1B-10)==0)
			c = 10;
		else if ((c=GSMtab[c+9])<32 && c!=13)
			c+= 0x390; /* some capital Latin letters second as Greek */
	}
	*s = p;

	return c;
}

const UTF16	__UTF16_MacRoman_high[128] =
{ /* NOTE: Apple(R) corp. char (0xF0) and euro/currency (0xDB) sign char vary */
/*80*/	196,197,199,201,209,214,220,225,
/*88*/	224,226,228,227,229,231,233,232,
/*90*/	234,235,237,236,238,239,241,243,
/*98*/	242,244,246,245,250,249,251,252,
/*A0*/	8224,176,162,163,167,8226,182,223,
/*A8*/	174,169,8482,180,168,8800,198,216,
/*B0*/	8734,177,8804,8805,165,181,8706,8721,
/*B8*/	8719,960,8747,170,186,937,230,248,
/*C0*/	191,161,172,8730,402,8776,8710,171,
/*C8*/	187,8230,160,192,195,213,338,339,
/*D0*/	8211,8212,8220,8221,8216,8217,247,9674,
/*D8*/	255,376,8260,8364,8249,8250,64257,64258,
/*E0*/	8225,183,8218,8222,8240,194,202,193,
/*E8*/	203,200,205,206,207,204,211,212,
/*F0*/	63743,210,218,219,217,305,710,732,
/*F8*/	175,728,729,730,184,733,731,711
};

__cb_utf32reader(readUTF32_MacRoman)
{
	UTF32		c;

	if ((c=*(*s)++)&0x80)
		c = __UTF16_MacRoman_high[c-0x80];
	return c;
}

const UTF16	__UTF16_WCP1250_high[128] =
{
/*80*/	8364,129,8218,131,8222,8230,8224,8225,
/*88*/	136,8240,352,8249,346,356,381,377,
/*90*/	144,8216,8217,8220,8221,8226,8211,8212,
/*98*/	152,8482,353,8250,347,357,382,378,
/*A0*/	160,711,728,321,164,260,166,167,
/*A8*/	168,169,350,171,172,173,174,379,
/*B0*/	176,177,731,322,180,181,182,183,
/*B8*/	184,261,351,187,317,733,318,380,
/*C0*/	340,193,194,258,196,313,262,199,
/*C8*/	268,201,280,203,282,205,206,270,
/*D0*/	272,323,327,211,212,336,214,215,
/*D8*/	344,366,218,368,220,221,354,223,
/*E0*/	341,225,226,259,228,314,263,231,
/*E8*/	269,233,281,235,283,237,238,271,
/*F0*/	273,324,328,243,244,337,246,247,
/*F8*/	345,367,250,369,252,253,355,729
};

__cb_utf32reader(readUTF32_WCP1250)
{
	UTF32		c;

	if ((c=*(*s)++)&0x80)
		c = __UTF16_WCP1250_high[c-0x80];
	return c;
}

const UTF16	__UTF16_WCP1252_high[32] =
{
	8364, 129, 8218, 402, /* euro, U+0081, low rsquo, f-hook */
	8222, 8230, 8224, 8225, /* low rdquo, hor-elipsis, dagger, Dagger */
	710, 8240, 352, 8249, /* circumflex, per mille, S-caron, angle lsquo */
	338, 141, 381, 143, /* OE ligature, U+008D, Z-caron, U+008F */
	144, 8216, 8217, 8220, /* U+0090, lsquo, rsquo, ldquo */
	8221, 8226, 8211, 8212, /* rdquo, bullet, en dash, em dash */
	732, 8482, 353, 8250, /* small tilde, trademark, s-caron, angle rsquo */
	339, 157, 382, 376 /* oe ligature, U+009D, z-caron, Y-umlaut */
};

/* reflected in Windows HTML browsers for ISO-8859-1 */
__cb_utf32reader(readUTF32_WCP1252)
{
	UTF32		c;

	if (((c=*(*s)++)-0x80)<=31)
		c = __UTF16_WCP1252_high[c-0x80];
	return c;
}

UTF16s		__UTF16_user8bit[256];
const UTF16	*__UTF16_userlast = 0;

/* may read the same character more than once for multi-character transcoding */
__cb_utf32reader(readUTF32_user8bit) /* not tested */
{
	const UTF16	*p = __UTF16_user8bit[**s];
	UTF32		c;

	if (__UTF16_userlast>p+3)
		__UTF16_userlast = p, c = *__UTF16_userlast++; /* start str */
	else if (c=*__UTF16_userlast++,__UTF16_userlast>p+3)
		goto pe; /* physical end of string */
	if (*__UTF16_userlast==0) /* terminated? */
	{ /* at end of string */
	pe:	__UTF16_userlast = 0, ++(*s);
	} else if (c-0xD800<=0x7FF)
		if ((c=c-0xD800<<10)>0x3FF<<10)
			c = UTF32_UNKNOWN|UTF32_ILLEGAL; /* valid UTF16 strs only! */
		else
		{
			unsigned	c2;

			if ((c2=*__UTF16_userlast++-0xDC00)>0x3FF)
				c2 = UTF32_ILLEGAL;
			c2+= 0x10000, c|= c2;
			if (*__UTF16_userlast==0)
				goto pe;
		}
	return c;
}

/* this routine parses 5/6-byte characters, but returns UNKNOWN for that case */
__cb_utf32reader(readUTF32_UTF8)
{
	const UTF8	*p = *s;
	UTF32		c;
	uchar		a;

	switch(c=*p++) /* switch it all for speed */
	{
	default: break; /* ASCII */

	/* sequence byte only - invalid */
	case 0x80: case 0x81: case 0x82: case 0x83: case 0x84: case 0x85:
	case 0x86: case 0x87: case 0x88: case 0x89: case 0x8A: case 0x8B:
	case 0x8C: case 0x8D: case 0x8E: case 0x8F: case 0x90: case 0x91:
	case 0x92: case 0x93: case 0x94: case 0x95: case 0x96: case 0x97:
	case 0x98: case 0x99: case 0x9A: case 0x9B: case 0x9C: case 0x9D:
	case 0x9E: case 0x9F: case 0xA0: case 0xA1: case 0xA2: case 0xA3:
	case 0xA4: case 0xA5: case 0xA6: case 0xA7: case 0xA8: case 0xA9:
	case 0xAA: case 0xAB: case 0xAC: case 0xAD: case 0xAE: case 0xAF:
	case 0xB0: case 0xB1: case 0xB2: case 0xB3: case 0xB4: case 0xB5:
	case 0xB6: case 0xB7: case 0xB8: case 0xB9: case 0xBA: case 0xBB:
	case 0xBC: case 0xBD: case 0xBE: case 0xBF:
	s0:	c = UTF32_UNKNOWN|UTF32_ILLEGAL;
		break;
	/* illegal, but used by some programs for C strings */
	case 0xC0: case 0xC1: c|= UTF32_ILLEGAL>>6;

	case 0xC2: case 0xC3: case 0xC4: case 0xC5: case 0xC6: case 0xC7:
	case 0xC8: case 0xC9: case 0xCA: case 0xCB: case 0xCC: case 0xCD:
	case 0xCE: case 0xCF: case 0xD0: case 0xD1: case 0xD2: case 0xD3:
	case 0xD4: case 0xD5: case 0xD6: case 0xD7: case 0xD8: case 0xD9:
	case 0xDA: case 0xDB: case 0xDC: case 0xDD: case 0xDE: case 0xDF:
		c-= 0xC0;
	s1:	c<<= 6;
		if ((a=*p++-0x80)>0x3F)
			{--p, c|= UTF32_ILLEGAL; break;}
		c|= a;
		break;
	case 0xED:
		if (*p>0x9F) /* UTF-16 surrogate range */
			c|= UTF32_ILLEGAL>>12;
		goto c2;
	case 0xE0:
		if (*p<0xA0)
			c|= UTF32_ILLEGAL>>12;
	case 0xE1: case 0xE2: case 0xE3: case 0xE4: case 0xE5: case 0xE6:
	case 0xE7: case 0xE8: case 0xE9: case 0xEA: case 0xEB: case 0xEC:
	case 0xEE: case 0xEF:
	c2:	c-= 0xE0;
	s2:	c<<= 6;
		if ((a=*p++-0x80)>0x3F)
			{--p, c|= UTF32_ILLEGAL; break;}
		c|= a;
		goto s1;
	case 0xF4:
		if (*p>0x8F)
			goto i3;
		goto c3;
	case 0xF0:
		if (*p<0x90)
			c|= UTF32_ILLEGAL>>18;
	case 0xF1: case 0xF2: case 0xF3:
	c3:	c = c-0xF0<<6;
		if ((a=*p++-0x80)>0x3F)
			{--p, c|= UTF32_ILLEGAL; break;}
		c|= a;
		goto s2;
	/* invalid sequences for the applicable UTF-32 range */
	case 0xFC: case 0xFD: case 0xFE: case 0xFF:
		if ((a=*p++-0x80)>0x3F)
			{--p; goto s0;}
	case 0xF8: case 0xF9: case 0xFA: case 0xFB:
		if ((a=*p++-0x80)>0x3F)
			{--p; goto s0;}
	case 0xF5: case 0xF6: case 0xF7:
	i3:	if ((a=*p++-0x80)>0x3F)
			{--p; goto s0;}
		if ((a=*p++-0x80)>0x3F)
			{--p; goto s0;}
		if ((a=*p++-0x80)>0x3F)
			--p;
		goto s0;
	}
	*s = p;

	return c;
}

__cb_utf32reader(readUTF32_UTF16LE) /* not tested */
{
	const uint8	*p = *s;
	UTF32		c, c2;

	c = *p++, c|= *p++<<8;
	if ((c-=0xD800)>0x3FF)
	{
		if (c<=0x7FF)
			c = UTF32_UNKNOWN|UTF32_ILLEGAL; /* low surrogate first */
	} else
	{
		c<<= 10; /* shift up high surrogate, and load low */
		c2 = *p++, c2|= *p++<<8;
		if ((c2-=0xDC00)>0x3FF) /* is low surrogate out of range? */
			c2 = UTF32_ILLEGAL, p-= 2; /* mark and discount */
		c2+= 0x10000, c|= c2;
	}
	*s = p;

	return c;
}

__cb_utf32reader(readUTF32_UTF16BE) /* not tested */
{
	const uint8	*p = *s;
	UTF32		c, c2;
#if TARGET_RT_BIG_ENDIAN
	c = *((const UTF16*)p)++; /* besides 68000/10, no BE CPUs will crash */
#else
	c = *p++<<8, c|= *p++;
#endif
	if ((c-=0xD800)>0x3FF)
	{
		if (c<=0x7FF)
			c = UTF32_UNKNOWN|UTF32_ILLEGAL; /* low surrogate first */
	} else
	{
		c<<= 10; /* shift up high surrogate, and load low */
#if TARGET_RT_BIG_ENDIAN
		c2 = *((const UTF16*)p)++; /* besides 68000/10, no BE CPUs will crash */
#else
		c2 = *p++<<8, c2|= *p++;
#endif
		if ((c2-=0xDC00)>0x3FF) /* is low surrogate out of range? */
			c2 = UTF32_ILLEGAL, p-= 2; /* mark and discount */
		c2+= 0x10000, c|= c2;
	}
	*s = p;

	return c;
}

__cb_utf32reader(readUTF32_UTF32LE) /* not tested */
{
	const uint8	*p = *s;
	UTF32		c;

	c = *p++, c|= *p++<<8, c|= *p++<<16, c|= *p++<<24;
	if (c-0xD800<0x800 || c>UTF32_MAX_LEGAL)
		c = UTF32_UNKNOWN|UTF32_ILLEGAL; /* UTF-16 surrogate */
	*s = p;

	return c;
}

__cb_utf32reader(readUTF32_UTF32BE) /* not tested */
{
	const uint8	*p = *s;
	UTF32		c;
#if TARGET_RT_BIG_ENDIAN
	c = *((const UTF32*)p)++; /* besides 68000/10, no BE CPUs will crash */
#else
	c = *p++<<8, c|= *p++, c<<= 16, c|= *p++<<8, c|= *p++;
#endif
	if (c-0xD800<0x800 || c>UTF32_MAX_LEGAL)
		c = UTF32_UNKNOWN|UTF32_ILLEGAL; /* UTF-16 surrogate */
	*s = p;

	return c;
}

/* ---- writers */

__cb_utf32writer(writeCharASCII)
{
	if (i>(UTF32)127)
		i = '?';
	*s = (uint8)i;
	return 1;
}

__cb_utf32writer(writeCharISO8859_1)
{
	if (i>(UTF32)255)
		i = '?';
	*s = (uint8)i;
	return 1;
}

__cb_utf32writer(writeCharGSM0338)
{
	const uint8	*hiTab = (const uint8*)"\x1B??\1$\3?_" "`"
	/*C0*/	"????[\x0E\x1C??\x1F??????"
	/*D0*/	"?]????\\?\x0B???^??\x1E"
	/*E0*/	"\x7F???{\x0F\x1D\x09\4\5??\7???"
	/*F0*/	"?}\x08???|?\x0C\6??~";
	const uint8	*escTab = (const uint8*)"</>\x14(@)=";
	const uint8	*GreekTab = (const uint8*)
		"AB\x13\x10EZH\x19IK\x14MN\x1AO\x16P?\x18TU\x12X\x17\x15";
	if (i>125)
	{
		if (i<0xFD)
			if ((i-=160)<8)
				{*s = hiTab[i]; goto end;}
			else if ((i-=0xBF-0xA3)<=0xFC-0xBF)
				{*s = hiTab[i+8]; goto end;}
		if ((i-=913)<=24)
			{*s = GreekTab[i]; goto end;}
		if ((i-=945-913)<=24)
			{*s = GreekTab[i]; goto end;}
		if ((i-=8364-945)==0)
			{i = 'e'; goto esc;}
	} else if (i<='@')
	{
		if ((i&~64)==0)
			{*s = 0; goto end;}
		if (i<32)
		{
			if ((i-=10)==0)
				{*s = 10; goto end;}
			else if ((i-=12-10)==0)
			{
				i = 10;
			esc:	*s++ = 0x1B, *s = i;
				return 2;
			} else if (--i==0)
				{*s = 13; goto end;}
		} else
		{
			if (i=='$')
				i = 2;
			goto set;
		}
	} else if (i<91)
		goto set;
	else if (i<97)
	{
		if ((i-='[')<='^'-'[')
			{i = escTab[i]; goto esc;}
		else if ((i-='{'-'[')<='}'-'{')
			{i = escTab[i+4]; goto esc;}
	} else if (i<123)
		{set: *s = i; goto end;}
unk:	*s = '?';
end:	return 1;
}

//#define MacRomanUNKNOWN 0x13 /* "System" black diamond (similar look to UNKNOWN) */

__cb_utf32writer(writeCharMacRoman)
{
	*s = (uint8)BMPwchr(i,__UTF16_MacRoman_high,128,128);
	return 1;
}

__cb_utf32writer(writeCharWCP1250)
{
	*s = (uint8)BMPwchr(i,__UTF16_WCP1250_high,128,128);
	return 1;
}

__cb_utf32writer(writeCharWCP1252)
{
	if (i>159)
		*s = (uint8)i;
	else
		*s = (uint8)BMPwchr(i,__UTF16_WCP1252_high,128,32);
	return 1;
}

__cb_utf32writer(writeCharUTF8) /* implementation not tested */
{
	unsigned long	h, l;

	if (i<(UTF32)0x80)
	{
		*s = (UTF8)i;
		return 1;
	}
	if (i>UTF32_MAX_LEGAL || i-0xD800<=0x7FF)
		i = UTF32_UNKNOWN;
	l = i>>6, i|= 0x80, i&=~0x40;
	if (i<(UTF32)0x800)
	{
		l|= 0xC0;
		*s++ = (UTF8)l, *s = (UTF8)i;
		return 2;
	}
	h = l>>6, l|= 0x80, l&= 0xBF;
	if (i<(UTF32)0x10000)
	{
		h|= 0xE0;
		*s++ = (UTF8)h, *s++ = (UTF8)l, *s = (UTF8)i;
		return 3;
	} /* already clipped */
	*s++ = (UTF8)(h>>6|0xF0), h|= 0x80, h&= 0xBF;
	*s++ = (UTF8)h, *s++ = (UTF8)l, *s = (UTF8)i;
	return 4;
}

__cb_utf32writer(writeCharUTF16LE)
{
	UTF16		l;

	if (i>UTF32_MAX_LEGAL || i-0xD800<=0x7FF)
		i = UTF32_UNKNOWN;
	if (i<=0xFFFF)
		{*s++ = (uint8)i, *s++ = (uint8)(i>>8); return 2;}
	l = (UTF16)((i>>10)+0xD7FF);
	*s++ = (uint8)l, *s++ = (uint8)(l>>8);
	i&= 0x3FF, i|= 0xDC00;
	*s++ = (uint8)i, *s = (uint8)(i>>8);
	return 4;
}

__cb_utf32writer(writeCharUTF16BE)
{
	UTF16		l;

	if (i>UTF32_MAX_LEGAL || i-0xD800<=0x7FF)
		i = UTF32_UNKNOWN;
	if (i<=0xFFFF)
		{*s++ = (uint8)(i>>8), *s++ = (uint8)i; return 2;}
	l = (UTF16)((i>>10)+0xD7FF);
#if TARGET_RT_BIG_ENDIAN
	*((UTF16*)s)++ = l; /* besides 68000/10, no BE CPUs will crash */
#else
	*s++ = (uint8)(l>>8), *s++ = (uint8)l;
#endif
	i&= 0x3FF, i|= 0xDC00;
#if TARGET_RT_BIG_ENDIAN
	*s = i;
#else
	*s++ = (uint8)(i>>8), *s = (uint8)i;
#endif
	return 4;
}

__cb_utf32writer(writeCharUTF32LE)
{
	if (i>UTF32_MAX_LEGAL || i-0xD800<=0x7FF)
		i = UTF32_UNKNOWN;
	*s++ = (uint8)i, *s++ = (uint8)(i>>8),
		*s++ = (uint8)(i>>16), *s = (uint8)(i>>24);
	return 4;
}

__cb_utf32writer(writeCharUTF32BE)
{
	if (i>UTF32_MAX_LEGAL || i-0xD800<=0x7FF)
		i = UTF32_UNKNOWN;
#if TARGET_RT_BIG_ENDIAN
	*s = i; /* besides 68000/10, no BE CPUs will crash */
#else
	*s++ = (uint8)(i>>24), *s++ = (uint8)(i>>16),
		*s++ = (uint8)(i>>8), *s = (uint8)i;
#endif
	return 4;
}

/* ---- transcoder */

/* buffer for multiple chars per read */
#define BUFFER_SIZE		64
#define MAX_CHAR_SIZE		6

static long int ftranscode(FILE *f,
	long int nb,
	utf_32_reader inroutine,
	int (*outroutine)(UTF32))
{
	uint8		t[BUFFER_SIZE+MAX_CHAR_SIZE], *r, *e;
	long int	c;

	if (f==0 || inroutine==0 || outroutine==0)
		return 0; /* idiot proofing */
	for(c=0,e=r=t+BUFFER_SIZE; nb>0; ++c)
	{
		UTF32		i;

		if (r>=e)
		{ /* need more bytes */
			int		z;

			if (e!=t+BUFFER_SIZE)
				break; /* end */
			if ((z=e-r)!=0) /* move remaining to front, if any */
			{ /* small, manual copy; memmove would be slower */
				const uint8	*s = r;
				uint8		*d = t;
				int		x = z;

				do
					*d++ = *s++;
				while(--x!=0);
			}
		read:	if ((i=BUFFER_SIZE-c)>nb)
				i = nb;
			if ((z=fread(r=t+z,1,i,f))<=0)
				if (z=0,r==t)
					break; /* no data read */
			nb-= z;
			e = r+z, r = t;
			e[2] = e[1] = *e = 0; /* terminate in case of overrun */
		}
		if ((i=inroutine(&r))>UTF32_MAX_LEGAL)
			c|= 1UL<<(sizeof(long int)*8)-1, i&= 0xFFFFFF;
		outroutine(i);
	}

	return c;
}
