/*
 charsets.h

  Created: 2010.11.11
  Revised: 2010.12.23
*/

/* ---- typedefs */

typedef unsigned char	uint8;

typedef unsigned long	UTF32;
typedef unsigned short	UTF16;
typedef uint8		UTF8;

/* Unicode character map string type -- 1-4 characters, null terminated */
typedef UTF16		UTF16s[4];
typedef UTF32		UTF32s[4];

#define __cb_utf32reader(x)	UTF32 x(uint8 const **s)
#define __cb_utf32writer(x)	int x(UTF32 i, uint8 *s)

typedef __cb_utf32reader((*utf_32_reader)); /* return char., updates position */
typedef __cb_utf32writer((*utf_32_writer)); /* return number of bytes written */

/* ---- constants */

/* use the mask for transcoding if necessary; */
/* an error occured if high byte isn't zero */
#define UTF32_CHAR_MASK		(UTF32)0x001FFFFF
#define UTF32_MAX_LEGAL		(UTF32)0x0010FFFF
#define UTF32_UNKNOWN		(UTF32)0x0000FFFD
#define UTF32_ILLEGAL		(UTF32)0x40000000

/* ---- prototypes */

int getMIB_IR_number(unsigned x);

const char *getMIB_kind(unsigned x);

/* returns MIB enum number, -1=undefined, 0=default (no name/proc for either) */
/* most encoding readers are unimplemented; you may have to use user8bit */
const int getEncodingByIndex(uint8 x,
	const char **name, /* can be NULL */
	utf_32_reader *reader); /* can be NULL */

/* the file transcoder */
/* returns number of characters read -- sets top bit if any chars were illegal */
static long int ftranscode(struct FILE *f,
	long int nb, /* number of bytes to read */
	utf_32_reader inroutine,
	int (*outroutine)(UTF32)); /* can be NULL */

/* writers */

__cb_utf32writer(writeCharASCII);
__cb_utf32writer(writeCharISO8859_1);
__cb_utf32writer(writeCharUTF8);
__cb_utf32writer(writeCharUTF16LE);
__cb_utf32writer(writeCharUTF16BE);
__cb_utf32writer(writeCharUTF32LE);
__cb_utf32writer(writeCharUTF32BE);

__cb_utf32writer(writeCharGSM0338); /* ETSI */
__cb_utf32writer(writeCharMacRoman);
__cb_utf32writer(writeCharWCP1250);
__cb_utf32writer(writeCharWCP1252);

/* readers */

__cb_utf32reader(readUTF32_ASCII); /* clears bit #7 */
__cb_utf32reader(readUTF32_ISO8859_1); /* uint8 read */
__cb_utf32reader(readUTF32_UTF8);
__cb_utf32reader(readUTF32_UTF16LE);
__cb_utf32reader(readUTF32_UTF16BE);
__cb_utf32reader(readUTF32_UTF32LE);
__cb_utf32reader(readUTF32_UTF32BE);

__cb_utf32reader(readUTF32_GSM0338); /* ETSI */
__cb_utf32reader(readUTF32_MacRoman);
__cb_utf32reader(readUTF32_WCP1250);
__cb_utf32reader(readUTF32_WCP1252); /* HTML browser default */
__cb_utf32reader(readUTF32_user8bit);

extern const UTF16 __UTF16_MacRoman_high[128], __UTF16_WCP1250_high[128];

extern const UTF16 __UTF16_WCP1252_high[32]; /* 80-9F; 8859-1 extension */

/* for use with readUTF32_user8bit */
extern UTF16s __UTF16_user8bit[256]; /* 2KB */
extern const UTF16 *__UTF16_userlast; /* for string sequences; can save or clear */
