#if 0
C -o map2list map2list.o -s Lib=Main
Exit 0
#endif
/*
 map2list.c
  *basic* character mapping to array printer

  Created: 2010.12.16 18:42 EST
  Prelimd: 2010.12.17 02:14 EST - yes, it took that long for such a basic executable
  Revised: 2010.12.21 00:24 EST - typo fix

Revised: 2010.12.17 23:12 EST - cosmetic fixes, putlong implementation
*/

#include <stdio.h>
#include <stdlib.h>

#define NO_VALUE		(long)0x80000000

/* for fast number printing - returns number of characters written */
int putlong(long i)
{
	unsigned char	n[12], *s = n+12;
	unsigned long	v = 0, c = 0;

	if ((v=i)&(long)0x80000000)
		putchar('-'), c = 1, v =-v;
	do
		v/= 10, *--s = i-v*10+'0', ++c;
	while((i=v)!=0);

	i = c;
	do
		putchar(*s++);
	while(--i!=0);

	return c;
}

/* returns next character/EOF; ~(next char.)-1 on error */
int freadlong(FILE *f,
	long *v)
{
	int		c;
	unsigned long	r = 0, m = 0;

	if (f==NULL) /* idiot proofing */
		{c = -1; goto empty;}
	if (v) /* initialize */
		*v = NO_VALUE;
	if ((c=fgetc(f))<=0)
		return c; /* sudden termination - valid */
	if ((c+=-'-')==0)
	{
		m = -1;
		if ((c=fgetc(f))<=0)
			goto empty; /* negative-empty */
	} else if ((c+='-'-'+')==0)
	{
		if ((c=fgetc(f))<=0)
			goto empty; /* positive-empty */
	} else if ((unsigned)((c+='+')-'0')>9)
		return c; /* non-numeric - valid */
	if (c=='0')
	{
		if ((c=fgetc(f))<=0)
			goto end; /* single zero */
		if (c!='x' && c!='X')
		{ /* octal */
			do
			{
				int	i = c;

				if ((unsigned)(i+=-'0')>7)
					if ((unsigned)i<=9
							|| (unsigned)(i+='0'-'A')<=('Z'-'A')
							|| (unsigned)(i+='A'-'a')<=('z'-'a')
							|| (i+='a'-'_')==0)
						return ~c-1; /* alphanumeric out of range */
					else
						break;
				r = r*8+i;
			} while((c=fgetc(f))>0);
		} else
		{ /* hexadecimal */
			if ((c=fgetc(f))<=0)
			{
			empty:
				return c ? -1:-2;
			}
			do
			{
				int	i = c;

				if ((unsigned)(i+=-'0')<=9)
					r = r*16+i;
				else if ((unsigned)(i+='0'-'A')<=('Z'-'A')
						|| (unsigned)(i+='A'-'a')<=('z'-'a'))
					if ((unsigned)i>('F'-'A'))
						return ~c-1; /* alphanumeric out of range */
					else
						r = r*16+i+10;
				else if ((i+='a'-'_')==0)
					return ~c-1;
				else
					break;
			} while((c=fgetc(f))>0);
		}
	} else
	{ /* decimal */
		do
		{
			int	i = c;

			if ((unsigned)(i+=-'0')>9)
				if ((unsigned)(i+='0'-'A')<=('Z'-'A')
						|| (unsigned)(i+='A'-'a')<=('z'-'a')
						|| (i+='a'-'_')==0)
					return ~c-1; /* alphanumeric out of range */
				else
					break;
			r = r*10+i;
		} while((c=fgetc(f))>0);
	}
end:
	if (v)
		*v = ((long int)r^m)-m; /* set applying sign */

	return c;
}

/* returns next character/EOF, <EOF on error */
int parseline(FILE *f,
	int p, /* previous character (mainly for line endings) */
	unsigned *x, /* index */
	int *r) /* remap */
{
	long int	v;
	int		c;

	if (x)
		*x = NO_VALUE;
	if (r)
		*r = NO_VALUE;
	while((c=freadlong(f,&v))>=-1)
	{
		if (v!=NO_VALUE)
			if (x) /* by way of exhausting the parameters */
				*x = v, x = 0;
			else if (r)
				*r = v, r = 0;
			else
				return -3;
		if (c<0)
			break;
		if (c==' ' || c=='\t') /* skip white space */
			continue;
		if (c=='\12') /* LF */
			if (p=='\15') /* DOS line ending; ignore */
				continue;
			else if (x || r)
				return -3;
			else
				break;
		if ((unsigned)(c-'\13')<='\15'-'\13') /* VT/FF/CR */
			break;
		if (c=='#') /* comment - skip line */
			while((c=fgetc(f))>=0)
				if ((unsigned)(c-'\12')<='\15'-'\12') /* LF/VT/FF/CR */
					return c;
		return -2; /* this character should not be here */
	}

	return c;
}

#define MAX_COUNT		sizeof(map)/sizeof(map[0])

int dumpmap(const char *n, /* file name */
	const int *r) /* range (-s and -e) */
{
	int		map[256];
	int		v, ec = 0, l, c;
	unsigned	x;
	FILE		*f;

	if ((f=fopen(n,"rb"))==0)
	{
		fprintf(stderr,"### Failed to open '%s'.\n",
			n);
		return 1;
	}

	/* default the list (one for one) */
	{
		int	*m = map, x = 0;

		do
			*m++ = x;
		while(++x<MAX_COUNT);
	}

	fprintf(stderr,"# Reading '%s'...\n",
		n);

	for(l=c=0; (c=parseline(f,c,&x,&v))!=-1;)
		if (++l,c<0 || (unsigned)x>=MAX_COUNT)
		{
		//	if (x==0xFFFD) /* "unknown" char */
		//		continue; /* ignore */
			if (x!=NO_VALUE && ++ec<=8)
			{
				const char	*s;

				if (c==-2)
					s = " - syntax error";
				else if (c==-3)
					s = " - expecting two parameters";
				else if (c<0)
					s = " - unknown token '%c'", c = ~(c+1);
				else if (x>=MAX_COUNT)
					s = " - 8-bit index out of range";
				else
					s = "";
				fprintf(stderr,"## Error at line #%d%s.\n",
					l,s,(char)c);
				if (ec>9)
				{
					fputs("### Too many errors.  Aborting.",stderr);
					break;
				}
			}
		} else if (v!=NO_VALUE)
			map[x] = v;
	/* close and print */
	fclose(f);
	fputs("# Closed.",stderr);

	/* print if range is valid (useful for predictable ranges, syntax checking) */
	if (r[0]<(int)(MAX_COUNT))
	{
		const int	*m = map;
		int		x, e = MAX_COUNT-1, s;

		putchar('\n'), putchar('{');

		if ((x=r[0])>0)
		{
			m+= x;
			if (x&7) /* uneven length; manually add indented line */
			{
				putchar('\n');
				if (r[0]>=0)
					putchar('/'), putchar('*'),
					putchar(x>0x9F ? (x>>4)+('A'-10):(x>>4)+'0'),
					putchar((x&15)>9 ? (x&15)+('A'-10):(x&15)+'0'),
					putchar('*'), putchar('/');
				putchar('\t');
			}
		} else
			x = 0;
		if (r[1]<MAX_COUNT)
			e = r[1];
		s = x;
		do
		{
			if (x!=s)
				putchar(',');
			if ((x&7)==0)
			{
				putchar('\n');
				if (r[0]>=0)
					putchar('/'), putchar('*'),
					putchar(x>0x9F ? (x>>4)+('A'-10):(x>>4)+'0'),
					putchar(x&8 ? '8':'0'),
					putchar('*'), putchar('/');
				putchar('\t');
			} else if (*m<=99) /* no space if >2 digits */
				putchar(' ');
		} while(putlong(*m++),++x<=e);

		putchar('\n'), putchar('}');
	}
	putchar('\n');

	return 0;
}

int main(int argc,
	char const*const* argv)
{
	int		r[2];
	char const	*const*a, *s;

	if (argc<=0 || argv==0)
		return -1;
	a = argv+1;
	if (--argc==0 || (s=*a)==0)
	{
		fprintf(stderr,"### Usage: %s [options] mapfile\n",
			argv[0]);
		return 1;
	}

	for(r[0]=r[1]=-1; --argc!=0 && *s=='-'; s=*++a)
		if (s[1]=='s' && !s[2])
			r[0] = atoi(*++a);
		else if (s[1]=='e' && !s[2])
			r[1] = atoi(*++a);
		else if ((unsigned char)(s[1]-'0')<=(unsigned char)9
				|| (s[1]=='m') && !s[2])
		{
			fprintf(stderr,
"### %s - multi-byte maps not supported by this version.\n",
				argv[0]);
			return 1;
		} else
		{
			fprintf(stderr,"### %s - unrecognized option '%s'.\n",
				argv[0],s);
			return 1;
		}
#if TARGET_OS_MAC
	if (((volatile char*)0x174)[60>>3]&(0x80>>(60&7))) /* keymap - control */
		Debugger();
#endif
	return dumpmap(s,r);
}
